"""
Acquired from https://github.com/quicksilver/Quicksilver
Which says

'''
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this software except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

'''

Apache License 2-0 is GPLv3+ compatible --ulrik


"""

__kupfer_name__ = _("Quicksilver Icons")
__description__ = ""
__version__ = ""
__author__ = "Quicksilver Authors. License: Apache 2.0"

import typing as ty

if ty.TYPE_CHECKING:
    from gettext import gettext as _  # noqa


def initialize_plugin(name):
    pass
